/* Lua/ysdl Awesome Tower Defense
 *
 * This is part of a Lua port of Awesome Tower Defense 0.3,
 * which is copyright (c) 2009 Trygve Vea and contributors
 * (read AUTHORS).
 *
 * This file copyright (C) 2010 Sean Bolton.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include "lua.h"
#include "lualib.h" 
#include "lauxlib.h" 

#include "SDL/SDL.h"

/* ---- stuff that should be in ysdl.h ---- */
static const char *ysdl_surface_name  = "ysdl_surface";
typedef struct _ysdl_surface {
    SDL_Surface *surface;
    int free;  /* true if we should free it */
} ysdl_surface;
void build_simple_class(lua_State *L, const char *name, const luaL_Reg *reg, int selfindex);
LUALIB_API int luaopen_ysdl (lua_State *L);
/* ---- ----- ---- ------ -- -- ------ ---- */

static const char *td_damage_array_name  = "td_damage_array";
static const char *td_map_name  = "td_map";

typedef struct _td_damage_array {
    int w, h;    /* width and height of array */
    int sw, sh;  /* surface width and height */
    int rw, rh;  /* width and height of damage rectangle */
    int has_damage;
    char     *data;
    SDL_Rect *rects;
} td_damage_array;

#define TD_MAP_MAX_STARTS  20
#define TD_MAP_ALGO_BOING   0
#define TD_MAP_ALGO_LEE     1

#define DIRECTION_N  0
#define DIRECTION_E  2
#define DIRECTION_S  4
#define DIRECTION_W  6

#define TD_MAP_TURN_N   1
#define TD_MAP_TURN_E   2
#define TD_MAP_TURN_S   3
#define TD_MAP_TURN_W   4
#define TD_MAP_MIRROR1  5
#define TD_MAP_MIRROR2  6
#define TD_MAP_TURN_R   7
#define TD_MAP_TURN_L   8
#define TD_MAP_RANDOM   9
#define TD_MAP_START1  10
#define TD_MAP_EXIT1   11
#define TD_MAP_START2  12
#define TD_MAP_EXIT2   13

typedef struct _td_map_cell {
    unsigned int traversable     : 1;  /* initial state of monster_allowed */
    unsigned int buildable       : 1;  /* initial state of tower_allowed */
    unsigned int monster_allowed : 1;
    unsigned int tower_allowed   : 1;
    unsigned int special         : 4;
    unsigned int tower           : 8;
} td_map_cell;

typedef struct _td_map_coord {
    short x, y;
} td_map_coord;

typedef struct _td_map {
    int            w, h;    /* width and height of array */
    int            sw, sh;  /* surface width and height */
    int            rw, rh;  /* width and height of each damage rectangle */
    td_map_cell   *map;
    unsigned char *dir;     /* pathfinding solution, directions are 0 to 7 clockwise from 12 o'clock */
    int            starts_count;
    td_map_coord   starts[TD_MAP_MAX_STARTS];
    int            starts_table_ref;
    int            exits_count;
    td_map_coord   exits[TD_MAP_MAX_STARTS];
    int            exits_table_ref;
    int            algorithm;
    void          *algo_data;
} td_map;

void _td_free(void *ptr) { free(ptr); }
void *_td_malloc(size_t size) { return malloc(size); } /* should do something responsible on failure */
void *_td_malloc0(size_t size) { return calloc(1, size); } /* likewise */

/* ==== td_damage_array ==== */

static int
td_damage_array__gc(lua_State *L)
{
    td_damage_array *a = lua_touserdata(L, 1);

    /* printf("td_damage_array__gc called for %p\n", a); */
    _td_free(a->data);
    _td_free(a->rects);

    return 0;
}

static int
td_damage_array_do_UpdateRects(lua_State *L)
{
    td_damage_array *a = luaL_checkudata(L, 1, td_damage_array_name);
    ysdl_surface *s = luaL_checkudata(L, 2, ysdl_surface_name);
    int x, xe, y, count;

    /* build list of damage SDL_Rects */
    count = 0;
    for (y = 0; y < a->h; y++) {
        for (x = 0; x < a->w; x++) {
            if (a->data[y * a->w + x]) {
                SDL_Rect *r;

                /* coallesce adjacent horizontal damage rectangles */
                xe = x + 1;
                while (xe < a->w && a->data[y * a->w + xe]) xe++;
                xe--;
                r = &a->rects[count++];
                r->x = x * a->rw;
                r->y = y * a->rh;
                r->w = (xe - x + 1) * a->rw;
                r->h = a->rh;

                x = xe;
            }
        }
    }
    if (count)
        SDL_UpdateRects(s->surface, count, a->rects);

    memset(a->data, 0, a->w * a->h);
    a->has_damage = 0;

    return 0;
}

static int
td_damage_array_get(lua_State *L)
{
    td_damage_array *a = luaL_checkudata(L, 1, td_damage_array_name);
    int x = lrint(floor(luaL_checknumber(L, 2)));  /* make sure to round down */
    int y = lrint(floor(luaL_checknumber(L, 3)));

    if (x < 0 || x >= a->w || y < 0 || y >= a->h)
        lua_pushnil(L);
    else
        lua_pushboolean(L, a->data[y * a->w + x]);

    return 1;
}

static int
td_damage_array_has_damage(lua_State *L)
{
    td_damage_array *a = luaL_checkudata(L, 1, td_damage_array_name);

    lua_pushboolean(L, a->has_damage);

    return 1;
}

static int
td_damage_array_set(lua_State *L)
{
    td_damage_array *a = luaL_checkudata(L, 1, td_damage_array_name);
    int x = lrint(floor(luaL_checknumber(L, 2)));  /* make sure to round down */
    int y = lrint(floor(luaL_checknumber(L, 3)));

    if (x >= 0 && x < a->w && y >= 0 && y < a->h) {
        a->data[y * a->w + x] = 1;
        a->has_damage = 1;
    }

    return 0;
}

static const luaL_Reg td_damage_array_lib[] = {
    { "__gc",            td_damage_array__gc },
    { "do_UpdateRects",  td_damage_array_do_UpdateRects },
    { "get",             td_damage_array_get },
    { "has_damage",      td_damage_array_has_damage },
    { "set",             td_damage_array_set },
    { NULL, NULL }
};

static int
td_new_damage_array(lua_State *L)
{
    int sw = luaL_checkinteger(L, 1);
    int sh = luaL_checkinteger(L, 2);
    int rw = luaL_checkinteger(L, 3);
    int rh = luaL_checkinteger(L, 4);
    td_damage_array *a;

    if (rw < 1 || rh < 1 || sw < rw || sh < rh) {
        lua_pushnil(L);
        return 1;
    }
    a = lua_newuserdata(L, sizeof(td_damage_array));     /* -- userdata */
    a->w = sw / rw;
    a->h = sh / rh;
    a->sw = sw;
    a->sh = sh;
    a->rw = rw;
    a->rh = rh;
    a->has_damage = 0;
    a->data = _td_malloc0(sizeof(char) * a->w * a->h);
    a->rects = _td_malloc0(sizeof(SDL_Rect) * a->w * a->h);
    /* printf("td_new_damage_array: %dx%d for %dx%d (%d,%d) at %p\n", a->w, a->h, sw, sh, rw, rh, a); */
    luaL_getmetatable(L, td_damage_array_name);          /* -- userdata metatable */
    lua_setmetatable(L, -2);                             /* -- userdata */
        
    return 1;
}

/* ==== Lee algorithm pathfinding ==== */

struct next_cell {
    int n;
    float cost;
    int dir;
};

struct lee_data {
    unsigned char *tmp;  /* temporary pathfinding map, 2x2 larger than original map */
    struct next_cell *next_array;
    int next_count;
    int next_count_max;
};

static void
lee_free(struct lee_data *l)
{
    _td_free(l->tmp);
    _td_free(l->next_array);
    _td_free(l);
}

static void
lee_add_next(struct lee_data *l, int n, float cost, int dir)
{
    int i, j;

    for (i = 0, j = 0; i < l->next_count; i++) {
        if (l->next_array[i].n == n)
            break;
        if (l->next_array[i].cost <= cost)
            j = i + 1;
    }
    if (i == l->next_count) {  /* not found in list */
        if (l->next_count == l->next_count_max) {  /* should not happen */
            fprintf(stderr, "ERROR: next_array is full!\n");
            exit(1);
        }
        if (j != l->next_count)  /* inserting in middle of list */
            memmove(&l->next_array[j + 1], &l->next_array[j], sizeof(struct next_cell) * (l->next_count - j));
        l->next_array[j].n    = n;
        l->next_array[j].cost = cost;
        l->next_array[j].dir  = dir;
        l->next_count++;
    } else {  /* found in list at index i */
        /* bias the path selection toward diagonal-first (so it looks slightly more intentional) */
        float bias = ((dir & 1) ? -0.1f : 0.0f) + ((l->next_array[i].dir & 1) ? 0.1f : 0.0f);
        if (l->next_array[i].cost > cost + bias) {
            if (j < i)
                memmove(&l->next_array[j + 1], &l->next_array[j], sizeof(struct next_cell) * (i - j));
            l->next_array[j].n    = n;
            l->next_array[j].cost = cost;
            l->next_array[j].dir  = dir;
        }
    }
}

#define ROOT2  1.4142135623731f

static int
lee_pathfind(td_map *m)
{
    int lw, lh;
    struct lee_data *l;
    td_map_cell *map;
    unsigned char *tmp;
    int xi, yi;

    /* allocations */
    lw = m->w + 2; lh = m->h + 2;  /* add space for padding around outside */
    l = (struct lee_data *)m->algo_data;
    if (l == NULL) {
        l = _td_malloc(sizeof(struct lee_data));
        m->algo_data = l;
        l->next_count_max = lw * 2 + lh * 2;
        l->tmp = (unsigned char *)_td_malloc(lw * lh);
        l->next_array = (struct next_cell *)_td_malloc(l->next_count_max * sizeof(struct next_cell));
    }
    map = m->map;
    tmp = l->tmp;
    l->next_count = 0;

    /* first, copy the maze, adding a solid obstacle border around it to
     * obviate edge checking */
    for (yi = 0; yi < lh; yi++) {
        for (xi = 0; xi < lw; xi++) {
            int n = yi * lw + xi;
            int s = (yi - 1) * m->w + (xi - 1);

            if (yi == 0 || yi == lh - 1 || xi == 0 || xi == lw - 1) {  /* added border cell */
                tmp[n] = 1;              /* mark as obstacle */
            } else {
                tmp[n] = map[s].monster_allowed ? 0 : 1;  /* copy open/obstacle from map */
            }
        }
    }

    /* next, initialize next_array with cells adjacent to exits */
    for (yi = 1; yi < lh - 1; yi++) {
        for (xi = 1; xi < lw - 1; xi++) {
            int n = yi * lw + xi;
            int s = (yi - 1) * m->w + (xi - 1);

            if (map[s].special == TD_MAP_EXIT1) {  /* or TD_MAP_EXIT2 */
                /* look for open adjacent cells and add them to next_array */
                int t;

                tmp[n] = 0x80;  /* mark as done */
                t = n + lw;
                if (tmp[t] == 0) lee_add_next(l, t, 1.0f, 0);
                t = n + lw - 1;
                if (tmp[t] == 0 && tmp[t - lw] == 0 && tmp[t + 1] == 0)
                    lee_add_next(l, t, ROOT2, 1);
                t = n - 1;
                if (tmp[t] == 0) lee_add_next(l, t, 1.0f, 2);
                t = n - lw - 1;
                if (tmp[t] == 0 && tmp[t + lw] == 0 && tmp[t + 1] == 0)
                    lee_add_next(l, t, ROOT2, 3);
                t = n - lw;
                if (tmp[t] == 0) lee_add_next(l, t, 1.0f, 4);
                t = n - lw + 1;
                if (tmp[t] == 0 && tmp[t + lw] == 0 && tmp[t - 1] == 0)
                    lee_add_next(l, t, ROOT2, 5);
                t = n + 1;
                if (tmp[t] == 0) lee_add_next(l, t, 1.0f, 6);
                t = n + lw + 1;
                if (tmp[t] == 0 && tmp[t - lw] == 0 && tmp[t - 1] == 0)
                    lee_add_next(l, t, ROOT2, 7);
            }
        }
    }

    /* process cells on next_array in order of cost, setting direction of least
     * cost, and adding open adjacent cells to next_array
     */
    while (l->next_count > 0) {
        int n = l->next_array[0].n;
        float cost = l->next_array[0].cost;
        int t;

        tmp[n] = 0x80 + l->next_array[0].dir;  /* mark it done */

        /* scan adjacent cells for open path, lee_add_next() them if they're open */
        t = n + lw;
        if (tmp[t] == 0) lee_add_next(l, t, cost + 1.0f, 0);
        t = n + lw - 1;
        if (tmp[t] == 0 && tmp[t - lw] == 0 && tmp[t + 1] == 0)
            lee_add_next(l, t, cost + ROOT2, 1);
        t = n - 1;
        if (tmp[t] == 0) lee_add_next(l, t, cost + 1.0f, 2);
        t = n - lw - 1;
        if (tmp[t] == 0 && tmp[t + lw] == 0 && tmp[t + 1] == 0)
            lee_add_next(l, t, cost + ROOT2, 3);
        t = n - lw;
        if (tmp[t] == 0) lee_add_next(l, t, cost + 1.0f, 4);
        t = n - lw + 1;
        if (tmp[t] == 0 && tmp[t + lw] == 0 && tmp[t - 1] == 0)
            lee_add_next(l, t, cost + ROOT2, 5);
        t = n + 1;
        if (tmp[t] == 0) lee_add_next(l, t, cost + 1.0f, 6);
        t = n + lw + 1;
        if (tmp[t] == 0 && tmp[t - lw] == 0 && tmp[t - 1] == 0)
            lee_add_next(l, t, cost + ROOT2, 7);

        /* remove top of next_array */
        memmove(&l->next_array[0], &l->next_array[1], sizeof(struct next_cell) * (l->next_count - 1));
        l->next_count--;
    }

    /* check for blockage: all start cells must connect to exits */
    for (xi = 0; xi < m->starts_count; xi++) {
        if (tmp[(m->starts[xi].y + 1) * lw + (m->starts[xi].x + 1)] < 0x80) {
            return 0; /* blocked */
        }
    }

    /* copy temporary solution into working solution */
    for (yi = 1; yi < lh - 1; yi++) {
        for (xi = 1; xi < lw - 1; xi++) {
            int n = yi * lw + xi;
            int s = (yi - 1) * m->w + (xi - 1);
#if 0 /* debug printout */
            char c;

            m->dir[s] = tmp[n] & 0x7f;
            switch (tmp[n]) {
              case 0:  c = ' '; break;
              case 1:  c = 'X'; break;
              case 2:  c = 'e'; break;
              default: if (tmp[n] >= 0x80) c = tmp[n] - 128 + '0'; else c = '?'; break;
            }
            printf("%c ", c);
        }
        printf("\n");
#else
            m->dir[s] = tmp[n] & 0x7f;
        }
#endif
    }

    return 1; /* not blocked */
}

/* ==== td_map ==== */

static int
td_map__gc(lua_State *L)
{
    td_map *m = lua_touserdata(L, 1);

    /* printf("td_map__gc called for %p\n", m); */
    _td_free(m->map);
    _td_free(m->dir);
    if (m->starts_table_ref != LUA_NOREF)
        luaL_unref(L, LUA_REGISTRYINDEX, m->starts_table_ref);
    if (m->exits_table_ref != LUA_NOREF)
        luaL_unref(L, LUA_REGISTRYINDEX, m->exits_table_ref);
    if (m->algo_data) {
        if (m->algorithm == TD_MAP_ALGO_LEE)
            lee_free(m->algo_data);
    }

    return 0;
}

static int
td_map_is_exit(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);
    int x = lrint(floor(luaL_checknumber(L, 2)));  /* make sure to round down */
    int y = lrint(floor(luaL_checknumber(L, 3)));

    if (x < 0 || x >= m->w || y < 0 || y >= m->h)
        lua_pushnil(L);
    else
        lua_pushboolean(L, m->map[y * m->w + x].special == TD_MAP_EXIT1); /* 'e' only, not 'E' for now */

    return 1;
}

static int
td_map_get_starts(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);

    if (m->starts_table_ref != LUA_NOREF) {
        lua_rawgeti(L, LUA_REGISTRYINDEX, m->starts_table_ref);
    } else {
        int i;

        lua_newtable(L);
        lua_pushvalue(L, -1);
        m->starts_table_ref = luaL_ref(L, LUA_REGISTRYINDEX);
        for (i = 0; i < m->starts_count; i++) {
            lua_newtable(L);
            lua_pushinteger(L, m->starts[i].x);  /* -- starttable subtable xcoord */
            lua_setfield(L, -2, "x");
            lua_pushinteger(L, m->starts[i].y);  /* -- starttable subtable ycoord */
            lua_setfield(L, -2, "y");            /* -- starttable subtable */
            lua_rawseti(L, -2, i + 1);           /* -- starttable */
        }
    }

    return 1;
}

static int
td_map_get_exits(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);

    if (m->exits_table_ref != LUA_NOREF) {
        lua_rawgeti(L, LUA_REGISTRYINDEX, m->exits_table_ref);
    } else {
        int i;

        lua_newtable(L);
        lua_pushvalue(L, -1);
        m->exits_table_ref = luaL_ref(L, LUA_REGISTRYINDEX);
        for (i = 0; i < m->exits_count; i++) {
            lua_newtable(L);
            lua_pushinteger(L, m->exits[i].x);  /* -- exittable subtable xcoord */
            lua_setfield(L, -2, "x");
            lua_pushinteger(L, m->exits[i].y);  /* -- exittable subtable ycoord */
            lua_setfield(L, -2, "y");           /* -- exittable subtable */
            lua_rawseti(L, -2, i + 1);          /* -- exittable */
        }
    }

    return 1;
}

static int
td_map_get_tower(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);
    int x = lrint(floor(luaL_checknumber(L, 2)));  /* make sure to round down */
    int y = lrint(floor(luaL_checknumber(L, 3)));

    lua_pushinteger(L, m->map[y * m->w + x].tower);

    return 1;
}

static int
td_map_monster_allowed(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);
    int x = lrint(floor(luaL_checknumber(L, 2)));  /* make sure to round down */
    int y = lrint(floor(luaL_checknumber(L, 3)));

    if (x < 0 || x >= m->w || y < 0 || y >= m->h)
        lua_pushnil(L);
    else
        lua_pushboolean(L, m->map[y * m->w + x].monster_allowed);

    return 1;
}

static int
td_map_monster_direction(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);
    int sx = luaL_checkinteger(L, 2);
    int sy = luaL_checkinteger(L, 3);
    int d = luaL_checkinteger(L, 4);  /* previous direction */
    int x = sx / m->rw;
    int y = sy / m->rh;

    if (x < 0 || x >= m->w || y < 0 || y >= m->h)
        lua_pushnil(L);
    else if (y == 0)
        lua_pushinteger(L, DIRECTION_S);
    else if (x == 0)
        lua_pushinteger(L, DIRECTION_E);
    else if (x == m->w - 1)
        lua_pushinteger(L, DIRECTION_W);
    else if (y >= m->h - 2)
        lua_pushinteger(L, DIRECTION_N);
    else if (m->algorithm == TD_MAP_ALGO_LEE) { /* use Lee algorithm results */
        lua_pushinteger(L, m->dir[y * m->w + x]);
    } else { /* do boing */
        unsigned int s = m->map[y * m->w + x].special;

        if (s > 0) {
            switch (s) {
              case TD_MAP_TURN_N: d = DIRECTION_N; break;
              case TD_MAP_TURN_E: d = DIRECTION_E; break;
              case TD_MAP_TURN_S: d = DIRECTION_S; break;
              case TD_MAP_TURN_W: d = DIRECTION_W; break;
              case TD_MAP_MIRROR1:  /* mirror S<->W or N<->E */
                switch (d) {
                  case DIRECTION_N: d = DIRECTION_E; break;
                  case DIRECTION_E: d = DIRECTION_N; break;
                  case DIRECTION_S: d = DIRECTION_W; break;
                  case DIRECTION_W: d = DIRECTION_S; break;
                }
                break;
              case TD_MAP_MIRROR2:  /* mirror S<->E or N<->W */
                switch (d) {
                  case DIRECTION_N: d = DIRECTION_W; break;
                  case DIRECTION_E: d = DIRECTION_S; break;
                  case DIRECTION_S: d = DIRECTION_E; break;
                  case DIRECTION_W: d = DIRECTION_N; break;
                }
                break;
              case TD_MAP_TURN_R: d = (d + 2) & 0x06; break;
              case TD_MAP_TURN_L: d = (d - 2) & 0x06; break;
              case TD_MAP_RANDOM: d = (d + (random() & 0x04) - 2) & 0x06; break;
            }
        }
        lua_pushinteger(L, d);
    }

    return 1;
}

static int
td_map_put_tower(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);
    int x = lrint(floor(luaL_checknumber(L, 2)));  /* make sure to round down */
    int y = lrint(floor(luaL_checknumber(L, 3)));
    int n = luaL_checkinteger(L, 4);
    td_map_cell *cell = &m->map[y * m->w + x];

    if (n == 0) {
        cell->monster_allowed = cell->traversable;
        cell->tower_allowed = cell->buildable;
        cell->tower = 0;
        if (m->algorithm == TD_MAP_ALGO_LEE)
            lee_pathfind(m);
        lua_pushboolean(L, 1);
    } else {
        cell->monster_allowed = 0;
        cell->tower_allowed = 0;
        cell->tower = n;
        if (m->algorithm == TD_MAP_ALGO_LEE) {
            if (!lee_pathfind(m)) { /* blocked */
                /* printf("blocked at %d, %d!\n", x, y); */
                cell->monster_allowed = cell->traversable;
                cell->tower_allowed = cell->buildable;
                cell->tower = 0;
                lua_pushboolean(L, 0);
            } else {
                lua_pushboolean(L, 1);
            }
        } else { /* not Lee, assume it can't block */
            lua_pushboolean(L, 1);
        }
    }

    return 1;
}

static int
td_map_reset(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);
    int x, y;

    for (y = 0; y < m->h; y++) {
        for (x = 0; x < m->w; x++) {
            td_map_cell *cell = &m->map[y * m->w + x];

            cell->monster_allowed = cell->traversable;
            cell->tower_allowed = cell->buildable;
            cell->tower = 0;
        }
    }
    if (m->algorithm == TD_MAP_ALGO_LEE)
        lee_pathfind(m);

    return 0;
}

static int
td_map_tower_allowed(lua_State *L)
{
    td_map *m = luaL_checkudata(L, 1, td_map_name);
    int x = lrint(floor(luaL_checknumber(L, 2)));  /* make sure to round down */
    int y = lrint(floor(luaL_checknumber(L, 3)));

    if (x < 0 || x >= m->w || y < 0 || y >= m->h)
        lua_pushnil(L);
    else
        lua_pushboolean(L, m->map[y * m->w + x].tower_allowed);

    return 1;
}

static const luaL_Reg td_map_lib[] = {
    { "__gc",              td_map__gc },
    { "is_exit",           td_map_is_exit },
    { "get_starts",        td_map_get_starts },
    { "get_exits",         td_map_get_exits },
    { "get_tower",         td_map_get_tower },
    { "monster_direction", td_map_monster_direction },
    { "monster_allowed",   td_map_monster_allowed },
    { "put_tower",         td_map_put_tower },
    { "reset",             td_map_reset },
    { "tower_allowed",     td_map_tower_allowed },
    { NULL, NULL }
};

static int
td_new_map(lua_State *L)
{
    int sw = luaL_checkinteger(L, 1);
    int sh = luaL_checkinteger(L, 2);
    int rw = luaL_checkinteger(L, 3);
    int rh = luaL_checkinteger(L, 4);
    const char *maptext = luaL_checkstring(L, 5);
    const char *pathfinding = luaL_checkstring(L, 6);
    int w, h, x, y;
    td_map *m;

    if (rw < 1 || rh < 1 || sw < rw * 2 || sh < rh * 3) {
        lua_pushnil(L);
        lua_pushliteral(L, "bad map dimensions");
        return 2;
    }
    w = sw / rw;
    h = sh / rh;
    if (strlen(maptext) != w * h) {
        lua_pushnil(L);
        lua_pushliteral(L, "map string length is incorrect");
        return 2;
    }
    m = lua_newuserdata(L, sizeof(td_map));     /* -- userdata */
    m->w = w;
    m->h = h;
    m->sw = sw;
    m->sh = sh;
    m->rw = rw;
    m->rh = rh;
    m->map = _td_malloc0(sizeof(td_map_cell) * w * h);
    m->dir = _td_malloc0(sizeof(unsigned char) * w * h);
    m->starts_count = 0;
    m->starts_table_ref = LUA_NOREF;
    m->exits_count = 0;
    m->exits_table_ref = LUA_NOREF;
    m->algorithm = (strcmp(pathfinding, "lee") == 0) ? TD_MAP_ALGO_LEE : TD_MAP_ALGO_BOING;
    m->algo_data = NULL;
    /* interpret and check map */
    for (y = 0; y < h; y++) {
        for (x = 0; x < w; x++) {
            unsigned char c = (unsigned char)*maptext++;
            unsigned int t, b, s;
            td_map_cell *cell = &m->map[y * w + x];

            if (x == 0 || x == w - 1 || y == 0 || y >= h - 2) { /* border */
                if (c != 's' && c != 'e' && c != 'S' && c != 'E' &&
                    c != '*' && c != 'X')
                    c = 'X'; /* force it to be a border character */
            }
            switch (c) {
              case '.':  t = 1; b = 1; s = 0;              break;
              case '+':  t = 0; b = 1; s = 0;              break;
              case ' ':  t = 1; b = 0; s = 0;              break;
              case '*':  t = 0; b = 0; s = 0;              break;
              case '^':  t = 1; b = 0; s = TD_MAP_TURN_N;  break;
              case '>':  t = 1; b = 0; s = TD_MAP_TURN_E;  break;
              case 'v':  t = 1; b = 0; s = TD_MAP_TURN_S;  break;
              case '<':  t = 1; b = 0; s = TD_MAP_TURN_W;  break;
              case '/':  t = 1; b = 0; s = TD_MAP_MIRROR1; break;
              case '\\': t = 1; b = 0; s = TD_MAP_MIRROR2; break;
              case '(':  t = 1; b = 0; s = TD_MAP_TURN_R;  break;
              case ')':  t = 1; b = 0; s = TD_MAP_TURN_L;  break;
              case '?':  t = 1; b = 0; s = TD_MAP_RANDOM;  break;
              case 'X':  t = 0; b = 0; s = 0;              break;
              case 's':  t = 1; b = 0; s = TD_MAP_START1;  break;
              case 'e':  t = 1; b = 0; s = TD_MAP_EXIT1;   break;
              case 'S':  t = 1; b = 0; s = TD_MAP_START2;  break;
              case 'E':  t = 1; b = 0; s = TD_MAP_EXIT2;   break;
              default:
                lua_pushfstring(L, "bad character in map at %d, %d", x, y);
                goto fail;
            }
            if (s == TD_MAP_START1 || s == TD_MAP_START2) { /* start */
                if (m->starts_count < TD_MAP_MAX_STARTS) {
                    m->starts[m->starts_count].x = x;
                    m->starts[m->starts_count].y = y;
                    m->starts_count++;
                } else {
                    lua_pushliteral(L, "map has too many starts");
                    goto fail;
                }
            } else if (s == TD_MAP_EXIT1 || s == TD_MAP_EXIT2) { /* exit */
                if (m->exits_count < TD_MAP_MAX_STARTS) {
                    m->exits[m->exits_count].x = x;
                    m->exits[m->exits_count].y = y;
                    m->exits_count++;
                } else {
                    lua_pushliteral(L, "map has too many exits");
                    goto fail;
                }
            }
            cell->traversable = t;
            cell->buildable   = b;
            cell->monster_allowed = cell->traversable;
            cell->tower_allowed = cell->buildable;
            cell->special = s;
            cell->tower = 0;
        }
    }
    if (m->algorithm == TD_MAP_ALGO_LEE) {
        if (!lee_pathfind(m)) {
            lua_pushliteral(L, "map is blocked!");
            goto fail;
        }
    }
    /* printf("td_new_map: %dx%d at %p\n", w, h, m); */
    /* set metatable for map */
    luaL_getmetatable(L, td_map_name);          /* -- userdata metatable */
    lua_setmetatable(L, -2);                    /* -- userdata */

    return 1;

  fail:
    _td_free(m->map);
    lua_pushnil(L);      /* -- userdata errmsg nil */
    lua_replace(L, -3);  /* -- nil errmsg */
    return 2;
}

/* ==== td functions ==== */

static const luaL_Reg main_lib[] = {
    { "new_damage_array",     td_new_damage_array },
    { "new_map",              td_new_map },
    { NULL, NULL }
};

/* ==== bit functions ==== */

static inline uint32_t
ubit_checkbitfield(lua_State *L, int index)
{
    /* assumes IEEE754 doubles, absolute value of argument < 2^51 */
    volatile union { double d; uint64_t i; } u;
    u.d = luaL_checknumber(L, index) + 6755399441055744.; /* + 2^52 + 2^51 */
    return (uint32_t)u.i;
}

static int
ubit_or(lua_State *L)
{
    uint32_t b = ubit_checkbitfield(L, 1);
    int i, n = lua_gettop(L);

    for (i = 2; i <= n; i++) b |= ubit_checkbitfield(L, i);

    lua_pushnumber(L, (double)b);

    return 1;
}

static const luaL_Reg ubit_lib[] = {
    { "bor", ubit_or },
    { NULL, NULL }
};

int
main (int argc, char *argv[])
{
    if (argc != 2) {
	fprintf (stderr, "usage: ysdl <ysdl-lua-script>\n");
	exit (1);
    }

    /* initialize Lua */ 
    lua_State *L = lua_open();
    luaL_openlibs(L);
    luaopen_ysdl(L);

    /* register main "td" functions */
    luaL_register(L, "td", main_lib);
    luaL_register(L, "ubit", ubit_lib);
    /* build object metatables */
    build_simple_class(L, td_damage_array_name, td_damage_array_lib, 1);
    build_simple_class(L, td_map_name,          td_map_lib, 1);

    lua_settop(L, 0);
	
    /* run the script */
    if (luaL_dofile(L, argv[1])) {
        const char *error = lua_tostring(L, -1);
        fprintf(stderr, "failed to load/run lua script '%s': %s\n", argv[1], error);
    }

    /* cleanup Lua */ 
    lua_close(L); 

    return 0;
}

