/* ysdl - libwhy lightweight Lua SDL binding
 *
 * Copyright (C) 2010 Sean Bolton.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/* Define these symbols to include the respective SDL extension
 * libraries: */
// #define YSDL_WANT_IMAGE
// #define YSDL_WANT_MIXER
// #define YSDL_WANT_TTF

#define YSDL_VERSION  "20100707"

#include <string.h>

#include "SDL/SDL.h"
#ifdef YSDL_WANT_IMAGE
#include "SDL/SDL_image.h"
#endif
#ifdef YSDL_WANT_MIXER
#include "SDL/SDL_mixer.h"
#endif
#ifdef YSDL_WANT_TTF
#include "SDL/SDL_ttf.h"
#endif

/* this has some meaning when building Windows DLLs: */
#define LUA_LIB

#include "lua.h"
#include "lauxlib.h"

static const char *ysdl_surface_name = "ysdl_surface";

typedef struct _ysdl_surface {
    SDL_Surface *surface;
    int free;  /* true if we should free it */
} ysdl_surface;

#ifdef YSDL_WANT_MIXER
static const char *ysdl_mix_chunk_name = "ysdl_mix_chunk";
static const char *ysdl_mix_music_name = "ysdl_mix_music";

typedef struct _ysdl_mix_chunk {
    Mix_Chunk *chunk;
} ysdl_mix_chunk;

typedef struct _ysdl_mix_music {
    Mix_Music *music;
} ysdl_mix_music;
#endif

#ifdef YSDL_WANT_TTF
static const char *ysdl_ttf_font_name = "ysdl_ttf_font";

typedef struct _ysdl_ttf_font {
    TTF_Font *font;
} ysdl_ttf_font;
#endif

void
build_simple_class(lua_State *L, const char *name, const luaL_Reg *reg, int selfindex)
{
    /* create metatable for object in the registry */
    luaL_newmetatable(L, name);                       /* -- metatable */
    /* register object metatable functions */
    luaL_register(L, NULL, reg);                      /* -- metatable */
    if (selfindex) {
        /* set __index to point to the metatable itself */
        lua_pushliteral(L, "__index");                /* -- metatable "__index" */
        lua_insert(L, -2);                            /* -- "__index" metatable */
        lua_rawset(L, -1);                            /* -- */
    } else {
        lua_pop(L, 1);                                /* -- */
    }
}

void *
make_simple_object(lua_State *L, size_t size, const char *class)
{
    void *u = lua_newuserdata(L, size);  /* -- userdata */

    luaL_getmetatable(L, class);         /* -- userdata metatable */
    lua_setmetatable(L, -2);             /* -- userdata */

    return u;
}

void
push_event_table(lua_State *L, SDL_Event *ev)  /* -DOCNOTE- */
{
    lua_createtable(L, 0, 4);  /* estimate */
    lua_pushinteger(L, ev->type);
    lua_setfield(L, -2, "type");              /* -- table */
    switch (ev->type) {
      case SDL_KEYDOWN:
        lua_pushinteger(L, ev->key.keysym.sym);
        lua_setfield(L, -2, "sym");
        lua_pushinteger(L, ev->key.keysym.mod);
        lua_setfield(L, -2, "mod");
        break;
      case SDL_MOUSEBUTTONDOWN:
        lua_pushinteger(L, ev->button.button);
        lua_setfield(L, -2, "button");
        lua_pushinteger(L, ev->button.x);
        lua_setfield(L, -2, "x");
        lua_pushinteger(L, ev->button.y);
        lua_setfield(L, -2, "y");
        break;
      case SDL_MOUSEMOTION:
        lua_pushinteger(L, ev->motion.x);
        lua_setfield(L, -2, "x");
        lua_pushinteger(L, ev->motion.y);
        lua_setfield(L, -2, "y");
        break;
    }
}

/* ==== SDL_Surface ==== */

static int
ysdl_surface__gc(lua_State *L)
{
    ysdl_surface *s = lua_touserdata(L, 1);

    /* printf("ysdl_surface__gc called for %p\n", s); */
    if (s->surface && s->free)
        SDL_FreeSurface(s->surface);
    s->surface = NULL;

    return 0;
}

static int
ysdl_DisplayFormat(lua_State *L)
{
    ysdl_surface *s0u = luaL_checkudata(L, 1, ysdl_surface_name);
    SDL_Surface *s1 = SDL_DisplayFormat(s0u->surface);

    if (s1 != NULL) {
        ysdl_surface *s1u = make_simple_object(L, sizeof(ysdl_surface), ysdl_surface_name);
        s1u->surface = s1;
        s1u->free = 1;  /* yes, we must free this */
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- SDL_Surface userdata | nil */
}

static int
ysdl_FillRect(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    SDL_Rect *r = lua_touserdata(L, 2);  /* note: no type checking of userdata, returns NULL for nil */
    Uint32 color = (Uint32)luaL_checknumber(L, 3);

    lua_pushinteger(L, SDL_FillRect(s->surface, r, color));

    return 1;
}

static int
ysdl_Flip(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);

    lua_pushinteger(L, SDL_Flip(s->surface));

    return 1;
}

static int
ysdl_FreeSurface(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);

    if (s->surface && s->free)
        SDL_FreeSurface(s->surface);
    s->surface = NULL;

    return 0;
}

static int
ysdl_GetClipRect(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    SDL_Rect *r;

    r = lua_newuserdata(L, sizeof(SDL_Rect));

    SDL_GetClipRect(s->surface, r);

    return 1;
}

static int
ysdl_SetAlpha(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    Uint32 flags = (Uint32)luaL_checknumber(L, 2);
    Uint8 alpha = luaL_checkinteger(L, 3);

    lua_pushinteger(L, SDL_SetAlpha(s->surface, flags, alpha));

    return 1;
}

static int
ysdl_SetColorKey(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    Uint32 flag = (Uint32)luaL_checknumber(L, 2);
    Uint32 key = (Uint32)luaL_checknumber(L, 3);

    lua_pushinteger(L, SDL_SetColorKey(s->surface, flag, key));

    return 1;
}

static int
ysdl_SetPalette1(lua_State *L)  /* -DOCNOTE- single-ncolor hack */
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    Uint32 flags = (Uint32)luaL_checknumber(L, 2);
    SDL_Color *colors = lua_touserdata(L, 3);  /* no type checking */
    int firstcolor = luaL_checkinteger(L, 4);

    lua_pushinteger(L, SDL_SetPalette(s->surface, flags, colors, firstcolor, 1));

    return 1;
}

static int
ysdl_UpdateRect(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    Sint32 x = luaL_checkinteger(L, 2);
    Sint32 y = luaL_checkinteger(L, 3);
    Sint32 w = luaL_checkinteger(L, 4);
    Sint32 h = luaL_checkinteger(L, 5);

    SDL_UpdateRect(s->surface, x, y, w, h);

    return 0;
}

/* -DOCNOTE-  surface:getPixel(x, y) returns pixel */
static int
ysdl_getPixel(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    SDL_Surface *surface = s->surface;
    int x = luaL_checkinteger(L, 2);
    int y = luaL_checkinteger(L, 3);
    /* public domain code shamelessly copied from the SDL documentation
     *   the surface must be locked before direct pixel access */
    int bpp = surface->format->BytesPerPixel;
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;
    Uint32 pixel;

    switch(bpp) {
    case 1: pixel = *p; break;
    case 2: pixel = *(Uint16 *)p; break;
    case 3: if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
                pixel = p[0] << 16 | p[1] << 8 | p[2];
            else
                pixel = p[0] | p[1] << 8 | p[2] << 16;
            break;
    case 4: pixel = *(Uint32 *)p; break;
    default: pixel = 0; break;
    }

    lua_pushnumber(L, (double)pixel);

    return 1;
}

static int
ysdl_getSize(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);

    lua_pushinteger(L, s->surface->w);
    lua_pushinteger(L, s->surface->h);

    return 2;
}

static const luaL_Reg ysdl_surfacelib[] = {
    { "__gc",          ysdl_surface__gc },
    { "DisplayFormat", ysdl_DisplayFormat },
    { "FillRect",      ysdl_FillRect },
    { "Flip",          ysdl_Flip },
    { "FreeSurface",   ysdl_FreeSurface },
    { "GetClipRect",   ysdl_GetClipRect },
    { "SetAlpha",      ysdl_SetAlpha },
    { "SetColorKey",   ysdl_SetColorKey },
    { "SetPalette1",   ysdl_SetPalette1 }, /* -DOCNOTE- */
    { "UpdateRect",    ysdl_UpdateRect },
    { "getPixel",      ysdl_getPixel },
    { "getSize",       ysdl_getSize },
    { NULL, NULL }
};

/* ==== SDL global functions ==== */

static int
ysdl_BlitSurface(lua_State *L)
{
    ysdl_surface *ss, *ds;
    SDL_Rect *sr, *dr;

    ss = luaL_checkudata(L, 1, ysdl_surface_name);
    sr = lua_touserdata(L, 2);  /* note: no type checking of userdata, returns NULL for nil */
    ds = luaL_checkudata(L, 3, ysdl_surface_name);
    dr = lua_touserdata(L, 4);  /* note: no type checking of userdata, returns NULL for nil */
    
    lua_pushinteger(L, SDL_BlitSurface(ss->surface, sr, ds->surface, dr));

    return 1;
}

static int
ysdl_CreateRGBSurface(lua_State *L)
{
    Uint32 flags = (Uint32)luaL_checknumber(L, 1);
    int width  = luaL_checkinteger(L, 2);
    int height = luaL_checkinteger(L, 3);
    int bitsPerPixel = luaL_checkinteger(L, 4);
    Uint32 Rmask = (Uint32)luaL_checknumber(L, 5);
    Uint32 Gmask = (Uint32)luaL_checknumber(L, 6);
    Uint32 Bmask = (Uint32)luaL_checknumber(L, 7);
    Uint32 Amask = (Uint32)luaL_checknumber(L, 8);
    SDL_Surface *s;

    s = SDL_CreateRGBSurface(flags, width, height, bitsPerPixel, 
                             Rmask, Gmask, Bmask, Amask);

    if (s != NULL) {
        ysdl_surface *u = make_simple_object(L, sizeof(ysdl_surface), ysdl_surface_name);
        u->surface = s;
        u->free = 1;  /* yes, we must free this */
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- SDL_Surface userdata | nil */
}

static int
ysdl_Delay(lua_State *L)
{
    Uint32 ms = (Uint32)luaL_checknumber(L, 1);

    SDL_Delay(ms);

    return 0;
}

static int
ysdl_GetError(lua_State *L)
{
    lua_pushstring(L, SDL_GetError());

    return 1;
}

static int
ysdl_GetKeyState(lua_State *L)  /* -DOCNOTE- takes a SDLK_* key number as argument, returns true if that key is pressed */
{
    static Uint8 *keystate = NULL;
    static int numkeys;
    int key = luaL_checkinteger(L, 1);

    if (keystate == NULL) keystate = SDL_GetKeyState(&numkeys);

    if (key < 0 || key >= numkeys)
        lua_pushnil(L);
    else
        lua_pushboolean(L, keystate[key]);

    return 1;
}

static int
ysdl_GetTicks(lua_State *L)
{
    lua_pushnumber(L, (double)SDL_GetTicks());

    return 1;
}

static int
ysdl_Init(lua_State *L)
{
    Uint32 flags = (Uint32)luaL_checknumber(L, 1);

    lua_pushinteger(L, SDL_Init(flags));

    return 1;
}

static int
ysdl_LoadBMP(lua_State *L)
{
    const char *filename = luaL_checkstring(L, 1);
    SDL_Surface *s = SDL_LoadBMP(filename);

    if (s != NULL) {
        ysdl_surface *u = make_simple_object(L, sizeof(ysdl_surface), ysdl_surface_name);
        u->surface = s;
        u->free = 1;  /* yes, we must free this */
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- SDL_Surface userdata | nil */
}

static int
ysdl_LowerBlit(lua_State *L)
{
    ysdl_surface *ss, *ds;
    SDL_Rect *sr, *dr;

    ss = luaL_checkudata(L, 1, ysdl_surface_name);
    sr = lua_touserdata(L, 2);  /* note: no type checking of userdata, returns NULL for nil */
    ds = luaL_checkudata(L, 3, ysdl_surface_name);
    dr = lua_touserdata(L, 4);  /* note: no type checking of userdata, returns NULL for nil */
    
    lua_pushinteger(L, SDL_LowerBlit(ss->surface, sr, ds->surface, dr));

    return 1;
}

static int
ysdl_MapRGB_using_Surface(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    Uint8 r = luaL_checkinteger(L, 2);
    Uint8 g = luaL_checkinteger(L, 3);
    Uint8 b = luaL_checkinteger(L, 4);
    lua_pushnumber(L, (double)SDL_MapRGB(s->surface->format, r, g, b));

    return 1;
}

static int
ysdl_MapRGBA_using_Surface(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    Uint8 r = luaL_checkinteger(L, 2);
    Uint8 g = luaL_checkinteger(L, 3);
    Uint8 b = luaL_checkinteger(L, 4);
    Uint8 a = luaL_checkinteger(L, 5);
    lua_pushnumber(L, (double)SDL_MapRGBA(s->surface->format, r, g, b, a));

    return 1;
}


static int
ysdl_PollEvent(lua_State *L)
{
    SDL_Event ev;
    int rc;

    rc = SDL_PollEvent(&ev);

    if (rc == 0)
        lua_pushnil(L);
    else
        push_event_table(L, &ev);

    return 1;
}

static int
ysdl_Quit(lua_State *L)
{
    SDL_Quit();

    return 0;
}

static int
ysdl_SaveBMP(lua_State *L)
{
    ysdl_surface *s = luaL_checkudata(L, 1, ysdl_surface_name);
    const char *filename = luaL_checkstring(L, 2);

    lua_pushinteger(L, SDL_SaveBMP(s->surface, filename));

    return 1;
}

static int
ysdl_SetVideoMode(lua_State *L)
{
    int width  = luaL_checkinteger(L, 1);
    int height = luaL_checkinteger(L, 2);
    int bitsperpixel = luaL_checkinteger(L, 3);
    Uint32 flags = (Uint32)luaL_checknumber(L, 4);
    SDL_Surface *s;

    s = SDL_SetVideoMode(width, height, bitsperpixel, flags);

    if (s != NULL) {
        ysdl_surface *u = make_simple_object(L, sizeof(ysdl_surface), ysdl_surface_name);
        u->surface = s;
        u->free = 0;  /* do not free this surface */
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- SDL_Surface userdata | nil */
}

static int
ysdl_WaitEvent(lua_State *L)
{
    SDL_Event ev;

    if (SDL_PollEvent(&ev) == 1)
        push_event_table(L, &ev);
    else
        lua_pushnil(L);

    return 1;
}

static int
ysdl_WM_SetCaption(lua_State *L)
{
    const char *title, *icon;

    if (lua_isnil(L, 1))
        title = NULL;
    else
        title = luaL_checkstring(L, 1);
    if (lua_isnil(L, 2))
        icon = NULL;
    else
        icon = luaL_checkstring(L, 2);
    SDL_WM_SetCaption(title, icon);

    return 0;
}

static int
ysdl_getRect(lua_State *L)  /* -DOCNOTE- Lau API addition: get values from a SDL_Rect userdata */
{
    SDL_Rect *r = lua_touserdata(L, 1);

    lua_pushinteger(L, r->x);
    lua_pushinteger(L, r->y);
    lua_pushinteger(L, r->w);
    lua_pushinteger(L, r->h);

    return 4;
}

static int
ysdl_newColor(lua_State *L)  /* -DOCNOTE- Lau API addition: create a SDL_Color userdata */
{
    Uint8 r = luaL_checkinteger(L, 1);
    Uint8 g = luaL_checkinteger(L, 2);
    Uint8 b = luaL_checkinteger(L, 3);
    SDL_Color *c;

    c = lua_newuserdata(L, sizeof(SDL_Color));
    c->r = r;
    c->g = g;
    c->b = b;
    /* don't need a __gc, Lua will just free the struct */

    return 1;
}

static int
ysdl_newRect(lua_State *L)  /* -DOCNOTE- Lau API addition: create a SDL_Rect userdata */
{
    int x, y, w, h;
    SDL_Rect *r;

    if (lua_gettop(L) > 0) {
        x = luaL_checkinteger(L, 1);
        y = luaL_checkinteger(L, 2);
    } else {
        x = 0;
        y = 0;
    }
    if (lua_gettop(L) > 2) {
        w = luaL_checkinteger(L, 3);
        h = luaL_checkinteger(L, 4);
    } else {
        w = 0;
        h = 0;
    }

    r = lua_newuserdata(L, sizeof(SDL_Rect));
    r->x = x;
    r->y = y;
    r->w = w;
    r->h = h;
    /* don't need a __gc, Lua will just free the struct */

    return 1;
}

static int
ysdl_setRect(lua_State *L)  /* -DOCNOTE- Lau API addition: update values in a SDL_Rect userdata */
{
    SDL_Rect *r = lua_touserdata(L, 1);
    int x = luaL_checkinteger(L, 2);
    int y = luaL_checkinteger(L, 3);
    int w = luaL_checkinteger(L, 4);
    int h = luaL_checkinteger(L, 5);

    r->x = x;
    r->y = y;
    r->w = w;
    r->h = h;

    return 0;
}

static int
ysdl_version(lua_State *L)
{
    lua_pushstring(L, YSDL_VERSION);

    return 1;
}

static const luaL_Reg sdllib[] = {
    { "BlitSurface",         ysdl_BlitSurface },
    { "CreateRGBSurface",    ysdl_CreateRGBSurface },
    { "Delay",               ysdl_Delay },
    { "GetError",            ysdl_GetError },
    { "GetKeyState",         ysdl_GetKeyState },
    { "GetTicks",            ysdl_GetTicks },
    { "Init",                ysdl_Init },
    { "LoadBMP",             ysdl_LoadBMP },
    { "LowerBlit",           ysdl_LowerBlit },
    { "MapRGB_using_Surface", ysdl_MapRGB_using_Surface }, /* -DOCNOTE- avoids having to deal with SDL_PixelFormat */
    { "MapRGBA_using_Surface", ysdl_MapRGBA_using_Surface }, /* -DOCNOTE- avoids having to deal with SDL_PixelFormat */
    { "PollEvent",           ysdl_PollEvent },
    { "Quit",                ysdl_Quit },
    { "SaveBMP",             ysdl_SaveBMP },
    { "SetVideoMode",        ysdl_SetVideoMode },
    { "WaitEvent",           ysdl_WaitEvent },
    { "WM_SetCaption",       ysdl_WM_SetCaption },
    { "newColor",            ysdl_newColor },
    { "getRect",             ysdl_getRect },
    { "newRect",             ysdl_newRect },
    { "setRect",             ysdl_setRect },
    { "ysdl_version",        ysdl_version },
    { NULL, NULL }
};

static const struct ysdl_enum_t {
    const char *name;
    int value;
} ysdl_enums[] = {
    { "AUDIO_U8",        AUDIO_U8 },   /* -DOCNOTE- */
    { "AUDIO_S16SYS",    AUDIO_S16SYS },   /* -DOCNOTE- */
    { "BIG_ENDIAN",      SDL_BIG_ENDIAN },
    { "BUTTON_LEFT",     SDL_BUTTON_LEFT },
    { "BYTEORDER",       SDL_BYTEORDER },
    { "HWSURFACE",       SDL_HWSURFACE },
    { "INIT_AUDIO",      SDL_INIT_AUDIO },
    { "INIT_EVERYTHING", SDL_INIT_EVERYTHING },
    { "INIT_TIMER",      SDL_INIT_TIMER },
    { "INIT_VIDEO",      SDL_INIT_VIDEO },
    { "K_1",             SDLK_1 },
    { "K_2",             SDLK_2 },
    { "K_3",             SDLK_3 },
    { "K_DOWN",          SDLK_DOWN },  /* -DOCNOTE- */
    { "K_ESCAPE",        SDLK_ESCAPE },
    { "K_LEFT",          SDLK_LEFT },
    { "K_RIGHT",         SDLK_RIGHT },
    { "K_SPACE",         SDLK_SPACE },
    { "K_UP",            SDLK_UP },
    { "K_e",             SDLK_e },
    { "K_f",             SDLK_f },
    { "K_p",             SDLK_p },
    { "K_q",             SDLK_q },
    { "K_r",             SDLK_r },
    { "K_s",             SDLK_s },
    { "K_t",             SDLK_t },
    { "K_w",             SDLK_w },
    { "KEYDOWN",         SDL_KEYDOWN },
    { "KEYUP",           SDL_KEYUP },
    { "LOGPAL",          SDL_LOGPAL },
    { "MOUSEBUTTONDOWN", SDL_MOUSEBUTTONDOWN },
    { "MOUSEBUTTONUP",   SDL_MOUSEBUTTONUP },
    { "MOUSEMOTION",     SDL_MOUSEMOTION },
    { "PRESSED",         SDL_PRESSED },
    { "QUIT",            SDL_QUIT },
    { "RLEACCEL",        SDL_RLEACCEL },
    { "SRCALPHA",        SDL_SRCALPHA },
    { "SRCCOLORKEY",     SDL_SRCCOLORKEY },
    { "SWSURFACE",       SDL_SWSURFACE },
    { "USEREVENT",       SDL_USEREVENT },
    { NULL, 0 }
};

#ifdef YSDL_WANT_IMAGE

/* ==== SDL_image (IMG) global functions ==== */

static int
ysdl_IMG_GetError(lua_State *L)
{
    lua_pushstring(L, IMG_GetError());

    return 1;
}

static int
ysdl_IMG_Load(lua_State *L)
{
    const char*filename = luaL_checkstring(L, 1);
    SDL_Surface *s = IMG_Load(filename);

    if (s != NULL) {
        ysdl_surface *u = make_simple_object(L, sizeof(ysdl_surface), ysdl_surface_name);
        u->surface = s;
        u->free = 1;  /* yes, we must free this */
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- SDL_Surface userdata | nil */
}


static const luaL_Reg imglib[] = {
    { "GetError",        ysdl_IMG_GetError },
    { "Load",            ysdl_IMG_Load },
    { NULL, NULL }
};

#endif  /* YSDL_WANT_IMAGE */

#ifdef YSDL_WANT_MIXER

/* ==== Mix_Chunk ==== */

static int
ysdl_Mix_Chunk__gc(lua_State *L)
{
    ysdl_mix_chunk *c = lua_touserdata(L, 1);

    /* printf("ysdl_Mix_Chunk__gc called for %p\n", c); */
    if (c->chunk)
        Mix_FreeChunk(c->chunk);
    c->chunk = NULL;

    return 0;
}

static int
ysdl_Mix_FreeChunk(lua_State *L)
{
    ysdl_mix_chunk *c = luaL_checkudata(L, 1, ysdl_mix_chunk_name);

    if (c->chunk)
        Mix_FreeChunk(c->chunk);
    c->chunk = NULL;

    return 0;
}

static const luaL_Reg ysdl_mix_chunklib[] = {
    { "__gc",             ysdl_Mix_Chunk__gc },
    { "FreeChunk",        ysdl_Mix_FreeChunk },
    { NULL, NULL }
};

/* ==== Mix_Music ==== */

static int
ysdl_Mix_Music__gc(lua_State *L)
{
    ysdl_mix_music *m = lua_touserdata(L, 1);

    /* printf("ysdl_Mix_Music__gc called for %p\n", m); */
    if (m->music)
        Mix_FreeMusic(m->music);
    m->music = NULL;

    return 0;
}

static int
ysdl_Mix_FreeMusic(lua_State *L)
{
    ysdl_mix_music *m = luaL_checkudata(L, 1, ysdl_mix_music_name);

    if (m->music)
        Mix_FreeMusic(m->music);
    m->music = NULL;

    return 0;
}

static const luaL_Reg ysdl_mix_musiclib[] = {
    { "__gc",             ysdl_Mix_Music__gc },
    { "FreeMusic",        ysdl_Mix_FreeMusic },
    { NULL, NULL }
};

/* ==== SDL_mixer (Mix) global functions ==== */

static int
ysdl_Mix_CloseAudio(lua_State *L)
{
    Mix_CloseAudio();

    return 0;
}

static int
ysdl_Mix_GetError(lua_State *L)
{
    lua_pushstring(L, Mix_GetError());

    return 1;
}

static int
ysdl_Mix_HaltChannel(lua_State *L)
{
    int channel = luaL_checkinteger(L, 1);

    lua_pushboolean(L, Mix_HaltChannel(channel));  /* "always returns zero. (kinda silly)" */

    return 1;
}

static int
ysdl_Mix_LoadMUS(lua_State *L)
{
    const char *file = luaL_checkstring(L, 1);
    Mix_Music *m = Mix_LoadMUS(file);

    if (m != NULL) {
        ysdl_mix_music *mu = make_simple_object(L, sizeof(ysdl_mix_music), ysdl_mix_music_name);
        mu->music = m;
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- Mix_Music userdata | nil */
}

static int
ysdl_Mix_LoadWAV(lua_State *L)
{
    const char *file = luaL_checkstring(L, 1);
    Mix_Chunk *c = Mix_LoadWAV(file);

    if (c != NULL) {
        ysdl_mix_chunk *cu = make_simple_object(L, sizeof(ysdl_mix_chunk), ysdl_mix_chunk_name);
        cu->chunk = c;
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- Mix_Chunk userdata | nil */
}

static int
ysdl_Mix_OpenAudio(lua_State *L)
{
    int frequency = luaL_checkinteger(L, 1);
    Uint16 format = luaL_checkinteger(L, 2);
    int channels = luaL_checkinteger(L, 3);
    int chunksize = luaL_checkinteger(L, 4);

    lua_pushinteger(L, Mix_OpenAudio(frequency, format, channels, chunksize));

#if 0
    // print music decoders available
    int i, max = Mix_GetNumMusicDecoders();
    printf("There are %d music decoders available\n", max);
    for(i=0; i<max; ++i)
        printf("Music decoder %d is for %s\n",i,Mix_GetMusicDecoder(i));
#endif

    return 1;
}

static int
ysdl_Mix_PlayChannel(lua_State *L)
{
    int channel = luaL_checkinteger(L, 1);
    ysdl_mix_chunk *c = luaL_checkudata(L, 2, ysdl_mix_chunk_name);
    int loops = luaL_checkinteger(L, 3);

    lua_pushinteger(L, Mix_PlayChannel(channel, c->chunk, loops));

    return 1;
}

static int
ysdl_Mix_Playing(lua_State *L)
{
    int channel = luaL_checkinteger(L, 1);

    lua_pushboolean(L, Mix_Playing(channel));

    return 1;
}

static int
ysdl_Mix_PlayingMusic(lua_State *L)
{
    lua_pushboolean(L, Mix_PlayingMusic());

    return 1;
}

static int
ysdl_Mix_PlayMusic(lua_State *L)
{
    ysdl_mix_music *m = luaL_checkudata(L, 1, ysdl_mix_music_name);
    int loops = luaL_checkinteger(L, 2);

    lua_pushinteger(L, Mix_PlayMusic(m->music, loops));

    return 1;
}

static const luaL_Reg mixlib[] = {
    { "CloseAudio",         ysdl_Mix_CloseAudio },
    { "GetError",           ysdl_Mix_GetError },
    { "HaltChannel",        ysdl_Mix_HaltChannel },
    { "LoadMUS",            ysdl_Mix_LoadMUS },
    { "LoadWAV",            ysdl_Mix_LoadWAV },
    { "OpenAudio",          ysdl_Mix_OpenAudio },
    { "PlayChannel",        ysdl_Mix_PlayChannel },
    { "Playing",            ysdl_Mix_Playing },
    { "PlayingMusic",       ysdl_Mix_PlayingMusic },
    { "PlayMusic",          ysdl_Mix_PlayMusic },
    { NULL, NULL }
};

#endif  /* YSDL_WANT_MIXER */

#ifdef YSDL_WANT_TTF

/* ==== TTF_Font ==== */

static int
ysdl_TTF_Font__gc(lua_State *L)
{
    ysdl_ttf_font *f = lua_touserdata(L, 1);

    /* printf("ysdl_TTF_Font__gc called for %p\n", f); */
    if (f->font)
        TTF_CloseFont(f->font);
    f->font = NULL;

    return 0;
}

static int
ysdl_TTF_CloseFont(lua_State *L)
{
    ysdl_ttf_font *f = luaL_checkudata(L, 1, ysdl_ttf_font_name);

    if (f->font)
        TTF_CloseFont(f->font);
    f->font = NULL;

    return 0;
}

static int
ysdl_TTF_RenderText_Solid(lua_State *L)
{
    ysdl_ttf_font *f = luaL_checkudata(L, 1, ysdl_ttf_font_name);
    const char *text = luaL_checkstring(L, 2);
    SDL_Color *fg = lua_touserdata(L, 3);  /* no type checking! */
    SDL_Surface *s = TTF_RenderText_Solid(f->font, text, *fg);

    if (s != NULL) {
        ysdl_surface *su = make_simple_object(L, sizeof(ysdl_surface), ysdl_surface_name);
        su->surface = s;
        su->free = 1;  /* yes, we must free this */
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- SDL_Surface userdata | nil */
}

static const luaL_Reg ysdl_ttf_fontlib[] = {
    { "__gc",             ysdl_TTF_Font__gc },
    { "CloseFont",        ysdl_TTF_CloseFont },
    { "RenderText_Solid", ysdl_TTF_RenderText_Solid },
    { NULL, NULL }
};

/* ==== SDL_ttf (TTF) global functions ==== */

static int
ysdl_TTF_GetError(lua_State *L)
{
    lua_pushstring(L, TTF_GetError());

    return 1;
}

static int
ysdl_TTF_Init(lua_State *L)
{
    lua_pushinteger(L, TTF_Init());

    return 1;
}

static int
ysdl_TTF_OpenFont(lua_State *L)
{
    const char *file = luaL_checkstring(L, 1);
    int ptsize = luaL_checkinteger(L, 2);
    TTF_Font *font = TTF_OpenFont(file, ptsize);

    if (font != NULL) {
        ysdl_ttf_font *fu = make_simple_object(L, sizeof(ysdl_ttf_font), ysdl_ttf_font_name);
        fu->font = font;
    } else {
        lua_pushnil(L);
    }

    return 1;  /* -- TTF_Font userdata | nil */
}

static int
ysdl_TTF_Quit(lua_State *L)
{
    TTF_Quit();

    return 0;
}

static const luaL_Reg ttflib[] = {
    { "GetError",        ysdl_TTF_GetError },
    { "Init",            ysdl_TTF_Init },
    { "OpenFont",        ysdl_TTF_OpenFont },
    { "Quit",            ysdl_TTF_Quit },
    { NULL, NULL }
};

#endif  /* YSDL_WANT_TTF */

/* ==== */

// const char *ysdl_init_string =
//     "SDL.INIT_EVERYTHING = 65535\n";

LUALIB_API int
luaopen_ysdl (lua_State *L)
{
    int i;
    int top = lua_gettop(L);

    /* register SDL libraries and build object metatables */
    luaL_register(L, "SDL", sdllib);                                   /* -- sdl */
    build_simple_class(L, ysdl_surface_name,   ysdl_surfacelib,  1);   /* -- sdl */
#ifdef YSDL_WANT_IMAGE
    luaL_register(L, "IMG", imglib);
    lua_pop(L, 1);                        /* 'SDL' is the table we'll return, so drop 'IMG' */
#endif
#ifdef YSDL_WANT_MIXER
    luaL_register(L, "Mix", mixlib);
    lua_pop(L, 1);
    build_simple_class(L, ysdl_mix_chunk_name, ysdl_mix_chunklib, 1);  /* -- sdl */
    build_simple_class(L, ysdl_mix_music_name, ysdl_mix_musiclib, 1);  /* -- sdl */
#endif
#ifdef YSDL_WANT_TTF
    luaL_register(L, "TTF", ttflib);
    lua_pop(L, 1);                                                     /* -- sdl */
    build_simple_class(L, ysdl_ttf_font_name,  ysdl_ttf_fontlib, 1);   /* -- sdl */
#endif

    /* register constants and lua functions */
    // if (luaL_dostring(L, ysdl_init_string) != 0)
    //     fprintf(stderr, "luaopen_ysdl error: %s\n", lua_tostring(L, -1));
    for (i = 0; ysdl_enums[i].name != NULL; i++) {
	lua_pushinteger(L, ysdl_enums[i].value);
	lua_setfield(L, -2, ysdl_enums[i].name);                       /* -- sdl */
    }

    /* make sure the stack is clean */
    if (top + 1 != lua_gettop(L))
        fprintf(stderr, "luaopen_ysdl warning: stack messiness!\n");

    return 1;                                                          /* -- sdl */
}

